import numpy as np
from copy import deepcopy

class NoAnnotation(BaseException):
    pass

class Segment(object):
    """ Temporal segment """
    def __init__(self, begin, end):
        self.begin = begin
        self.end = end
        assert(begin <= end)
        
    def intersection(self, s):
        if s.isempty():
            return EmptySegment()
            
        begin_int = max(self.begin, s.begin)
        end_int = min(self.end, s.end)
        if begin_int <= end_int:
            return Segment(begin_int, end_int)
        else:
            return EmptySegment()

    def union_length(self, s):
        li = self.intersection(s).length()
        return self.length() + s.length() - li
    
    def length(self):
        return self.end - self.begin

    def __eq__(self, s):
        return self.begin == s.begin and self.end == s.end
    
    def isempty(self):
        return self.begin >= self.end

    def __repr__(self):
        return "Segment(%s, %s)" % (self.begin, self.end)
    
    def __str__(self):
        return "[%s, %s)" % (self.begin, self.end)
    
    def totuple(self):
        return self.begin, self.end


class EmptySegment(Segment):
    def __init__(self):
        super(EmptySegment, self).__init__(0, 0)
        self.begin = None
        self.end = None
    
    def length(self):
        return 0
    
    def __eq__(self, s):
        return s.isempty()


class ImportanceMap(object):
    """ Importance map = temportal video segmentation + importance.
    Additionally contains a set of tags for each segment (shot_tags) and 
    for the whole video (video_tags).

    Usage example:
    # Set the segmentation by passing segment "begins" and "ends"
    x = ImportanceMap("video001", begins=[0, 200, 500], ends=[200, 500, 700])
    # Set importance of each segment
    x.importance = [0, 1.0/3, 1.0]
    # Set period 0.5 sec for the second segment
    x.shot_tags = ["", "prop:minlen=0.5s", ""]
    # Set the video fps
    x.video_tags = "prop:fps=30.000"
    # Print periods
    print x.get_periods()    # periods measured in frames: [200, 15, 200] 
    """

    def __init__(self, vidname, stub=None, begins=None, ends=None, debug=False):
        self.vidname = vidname
        assert stub == None
        if ends != None:
            ends = list(ends)
            if begins == None:
                begins = [0] + list(ends)[:-1]
            else:
                begins = list(begins)
            assert(len(begins) == len(ends))
            self.segments = [Segment(b,e) for b,e in zip(begins, ends)]
            self.importance = [0] * len(ends)
            self.shot_tags = ["" for i in range(len(ends))]
            self.video_tags = ""
        self.__periods = None
        self.__debug = debug
        
    def choose_segments(self, inds):
        """ Choose segments by list of indices and return as a new object."""
        x = ImportanceMap(self.vidname)
        x.segments = [deepcopy(self.segments[i]) for i in inds]
        x.importance = [self.importance[i] for i in inds]
        x.shot_tags = [deepcopy(self.shot_tags[i]) for i in inds]
        x.video_tags = deepcopy(self.video_tags)
        if self.__periods:
            x.__periods = [self.__periods[i] for i in inds]
        return x
    
    def argsort(self):
        """ Sort by decreasing importance,
        in case of equality, by increasing length
        in case of equality, by increasing starting time (for determinism). """
        sorting_array = zip(-np.array(self.importance),
            np.array(self.get_durations()),
            [s.begin for s in self.segments],
            range(len(self)))
        sorting_array.sort()
        
        from itertools import izip
        issorted = lambda l : all(a <= b for a,b in izip(l[:-1],l[1:]))
        assert(issorted(zip(*sorting_array)[0])), "Error in sorting..."
        
        order = zip(*sorting_array)[-1]
        return list(order)
    
    def choose_top_n(self, n_top, verbose=False):
        """ Choose top n segments and sort them in descending importance order.
            Return as a new object. """
        n_det = len(self.segments)
        if n_det < n_top:
            if verbose:
                print "WARN: only %d segments in output" % n_det
            n_top = n_det

        nans = np.isnan(self.importance)
        if np.any(nans):
            print "WARN: NaNs in self.importance!"
            for i in np.flatnonzero(nans):
                self.importance[i] = 0 
        
        order = self.argsort()
        order = np.array(order)
        
        if n_det > n_top and order[n_top] == order[n_top-1]:
            if verbose:
                print "WARN: top-%d segments are ambiguous" % n_top
        
        return self.choose_segments(order[:n_top])

    def check_contiguous(self):
        """ Check contiguous segmentation. """
        for i in range(len(self.segments)-1):
            s1 = self.segments[i]
            s2 = self.segments[i+1]
            assert(s1.end == s2.begin), "Segmentation non-contiguous"

    def merge_smaller(self, n_frames, method = "toleft"):
        """ Merge segments smaller than n_frames with a neighbour. """
        self.check_contiguous()

        if method == 'toleft':
            items = list(enumerate(self.segments))[::-1]
        else:
            items = list(enumerate(self.segments))

        valid = []
        # Variable s is a reference to object, so modification is applied 
        for i, s in items:
            if s.length() < n_frames:
                if i > 0 and (method == 'toleft' or i == len(self.segments)-1):
                    self.segments[i-1].end += s.length()
                else:
                    assert (i!=len(self.segments)-1), "assumed >= 2 segments"
                    self.segments[i+1].begin -= s.length()
            else:
                valid.append(i)
        valid = sorted(valid)
        self.segments = [self.segments[i] for i in valid]
        self.importance = [self.importance[i] for i in valid]
        self.shot_tags = [self.shot_tags[i] for i in valid]
        self.__periods = None

    def enlarge_smaller(self, min_frames):
        """ Extend segments smaller than min_frames to be exactly min_frames.
        Segmentation becomes non-contiguous. """
        self.__periods = None
        self.check_contiguous()
        n_frames_total = self.segments[-1].end
        for i, s in enumerate(self.segments):
            if s.length() < min_frames:
                ext = min_frames - s.length()
                if s.end + ext//2 + ext%2 > n_frames_total:
                    s.begin = n_frames_total - min_frames
                    s.end = n_frames_total
                else:
                    s.begin = max(0, s.begin-ext//2)
                    s.end = s.begin + min_frames
                assert s.begin >= 0
                assert s.end <= n_frames_total
        
        # Remove repeating segments
        i = 0
        while i < len(self.segments)-1:
            if self.segments[i] == self.segments[i+1]:
                self.segments.pop(i+1)
                self.importance.pop(i+1)
                self.shot_tags.pop(i+1)
            else:
                i += 1
    
    def decrease_importance(self, hinge):
        """ Linearly decrease importance of segments smaller than hinge. """
        for i, s in enumerate(self.segments):
            if s.length() < hinge:
                self.importance[i] *= (float(s.length()) / hinge)
    
    def choose_first_limited_duration(self, max_duration, method='cut'):
        """ Choose first segments that fit into the max_duration limit.
        max_duration - measured in *frames* """
        assert(method == 'cut')
        max_duration = int(np.round(max_duration))
        x = deepcopy(self)
        x.__periods = None
        duration = 0
        for i,s in enumerate(x.segments):
            if s.length() >= max_duration - duration:
                s.end = s.begin + max_duration - duration
                x.segments[i+1:] = []
                x.importance[i+1:] = []
                x.shot_tags[i+1:] = []
                break
            duration += s.length()
        return x

    def append(self, y):
        """ Assuming that the most important segments are not selected yet."""
        self.__periods = None

        # Fix different fps
        old_fps = y.get_fps()
        new_fps = self.get_fps()
        if new_fps != old_fps:
            for s in y.segments:
                old_len = s.length()/old_fps
                # Recompute segments with new framestep
                s.begin = int(float(s.begin)/old_fps*new_fps + 0.5)
                s.end = int(float(s.end)/old_fps*new_fps + 0.5)
                new_len = s.length()/new_fps
                assert(abs(old_len - new_len) < 1./new_fps + 1./old_fps)

        # Append to the end        
        duration = max(s.end for s in self.segments)
        y = deepcopy(y)
        for s in y.segments:
            s.begin += duration
            s.end += duration

        for f in ('segments', 'shot_tags', 'importance'):
            setattr(self, f, list(getattr(self, f)) + list(getattr(y, f)))
        self.vidname = '%s_%s' % (self.vidname, y.vidname)

    def top_limited_duration_inds(self, max_duration=None, method='flex',
        max_segment=None):
        """ Find indices of the top-scored segments that fit in the duration
        limit.
        max_duration - upper margin **in frames**
        method can be:
           'flex':   last segment is included even if it does not fit completely
           'strict': do not include segments that do not fit completely, but
                continue until there are segments that fit
        max_segment - upper margin for the duration of a single segment
        
        Returns indices of the corresonding segments.
        """
        assert(method in ['flex', 'strict'])
        order = self.argsort()
        top_pred_segm = []
        duration = 0
        pred_durations = self.get_durations()
        while duration < max_duration and order != []:
            i = order.pop(0)
            if max_segment and pred_durations[i] > max_segment:
                continue
            if method == 'strict' and \
                duration + pred_durations[i] > max_duration:
                continue
            top_pred_segm.append(i)
            duration += pred_durations[i]
        return top_pred_segm
                
    def __extract_periods(self):
        fps = self.get_fps()
        periods = []
        for i in range(len(self.shot_tags)):
            default_p = self.segments[i].length()
            p = None

            errors = False
            for t in self.shot_tags[i].split(' '):
                if t.startswith('prop:minlen='):
                    assert(p == None)
                    p = t.split('=')[1]
                    if p[-1] == 's':
                        p = int(float(p[:-1]) * fps)
                    else:
                        assert False, "Old period format"
                        p = int(p)
                    if not self.__debug:
                        assert(p <= default_p), self.vidname
                    else:
                        if p > default_p:
                            errors = True
            if self.__debug and errors:
                print "Error in periods"
            if p == None:
                p = default_p
            periods.append(p)
        self.__periods = periods
    
    def get_durations(self):
        """ Get segment durations. (in *frames*)"""
        return [s.length() for s in self.segments]
    
    # WARNING: changing periods not recommended
    def get_periods(self):
        """ Extract periods from tags.
        Periods duration is converted to *frames*.  """
        if self.__periods == None:
            self.__extract_periods()
        return self.__periods
    
    def get_fps(self):
        """ Extract fps from tags """
        fps = None
        for t in self.video_tags.split(' '):
            if t.startswith('prop:fps='):
                assert(fps is None), "Repeating fps occured"
                fps=float(t.split('=')[1])
        return fps

    def __len__(self):
        return len(self.segments)
    
    def __repr__(self):
        return self.to_dict().__repr__()
