import numpy as np
from copy import deepcopy

def metric_total_importance(gt_imap, pred_imap, n_top=5, max_duration=None, method='cut',
    alpha=0.8, verbose=True):
    """ Compute 1-to-many metric for 1 video using n_top detected segments.
    max_duration - maximum duration in *frames*.
    """
    
    if max_duration:
        max_duration = int(np.round(max_duration))
    
    gt_probs = gt_imap.importance
    n_gt = len(gt_probs)
    periods = gt_imap.get_periods()

    # Sort by decreasing importance, then increasing length and
    # select n_top if it is specified
    if n_top != None:
        pred_imap = pred_imap.choose_top_n(n_top, verbose=verbose)
    else:
        pred_imap = pred_imap.choose_top_n(len(pred_imap))
        
    duration = 0
    gt_covered = [0] * n_gt
    finish = False
    for ip, ps in enumerate(pred_imap.segments):
        if max_duration != None:
            if method == 'cut':
                # Take the largest part of the detected segment that fits in time
                largest_end = ps.begin + max_duration - duration
                if ps.end > largest_end:
                    ps.end = largest_end               
                    finish = True
            else:
                assert False
        duration += ps.length()
        
        for ig, gs in enumerate(gt_imap.segments):
            if gt_covered[ig]:
                continue

            assert(periods[ig] <= gs.length()), "Mistake in annotation"
            
            inter = ps.intersection(gs)
            if inter.length() > 0:
                if inter.length() > alpha * periods[ig]:
                    gt_covered[ig] = 1

        # Assuming integer segment boundaries
        if finish:
            assert(duration == max_duration)
        if duration == max_duration:
            break

    # Accumulate total importance
    summary_importance = 0
    for ig in range(n_gt):
        if gt_covered[ig]:
            summary_importance += gt_imap.importance[ig]

    return summary_importance, duration


def metric_importance_ratio(gt_imap, pred_imap, alpha=0.8, **kwargs):
    """ Metric normalized by maximum (gt) importance.
    Assuming max_duration set, n_top = None.
    
    imap's are not changed
    """
    assert(kwargs['n_top'] == None)
    assert(kwargs['max_duration'] != None)
    assert(kwargs.get('method', 'cut') == 'cut')
    
    unnorm_imp, dur = metric_total_importance(gt_imap, pred_imap, alpha=alpha, **kwargs)
   
    if (0 in gt_imap.get_durations()) or (0 in pred_imap.get_durations()):
        raise ValueError("Zero segment in annotation")

    # Construct ideal summary
    ideal = gt_imap.choose_top_n(len(gt_imap))
    periods = ideal.get_periods()
    for i,s in enumerate(ideal.segments):
        s.end = s.begin + int(np.ceil(1.0001*alpha*periods[i]))
        # indeed this is not needed as periods of pred_imap do not matter
        periods[i] = s.end - s.begin

    max_imp, _ = metric_total_importance(gt_imap, ideal, alpha=alpha, **kwargs)

    if unnorm_imp > max_imp:
        pass
        #print "WARN: maximum importance is suboptimal"
        
    if max_imp > 0:
        imp = float(unnorm_imp) / float(max_imp)
    else:
        imp = 1.0

    return imp, dur


def _insert_and_merge_segment(summary, s):
    """ Insert a segment in the summary.
    Merge if it adjacent to some other segment.
    Assuming: 1) summary is already in a sorted and merged state
    2) intersection of s and any segment in imap is of length 0"""

    n = len(summary)
    for i in range(n):
        assert(s.intersection(summary[i]).length() == 0), "Intersecting."
        if i > 0:
            assert(summary[i-1].end < summary[i].begin), "Not merged summary."

    for i in range(n):
        cs = summary[i]
        if s.end < cs.begin:
            summary.insert(i, deepcopy(s))
            return s
        elif s.end == cs.begin:
            cs.begin = s.begin
            return cs
        elif cs.end == s.begin:
            cs.end = s.end
            if i < n-1 and cs.end == summary[i+1].begin:
                cs.end = summary[i+1].end
                del summary[i+1]
            return cs
    if n == 0 or summary[-1].end < s.end:
        summary.append(deepcopy(s))
        return s
    else:
        assert False

def metric_msd(gt_imap, pred_imap):
    """ Compute Meaningful Summary Duration (aka Required Summary Size).
    See paper for definition.
    In our experiments, overall value for a set of videos with multiple
    annotations is the median over the flattened list of annotations.
    I.e. we take the median of all the scores (only once, no averaging per
    video).
    """
    alpha = 0.8

    # Sort the segments in the reverse order in case of equal
    # importance and period ("Users" case).
    # The opposite order gives the same whole picture.
    tuples = [(-x,y,-z,i) for i, (x,y,z) in enumerate(zip(pred_imap.importance,
        pred_imap.get_periods(), [a.begin for a in pred_imap.segments]))]
    tuples = sorted(tuples)
    i_segments = [x[3] for x in tuples]
    segments = [pred_imap.segments[i] for i in i_segments]

    max_importance = np.max(gt_imap.importance)

    cur_duration = 0.0
    fps = pred_imap.get_fps()
    gt_periods = gt_imap.get_periods()
    summary = []

    for ps, pp in zip(segments, pred_imap.get_periods()):
        cur_duration += ps.length() / fps
        # Segment may expand because of merging with the summary
        ps2 = _insert_and_merge_segment(summary, ps)
        for i, gs in enumerate(gt_imap.segments):
             inter = ps2.intersection(gs)
             if inter.length() > 0:
                 if inter.length() > alpha * gt_periods[i] and \
                 abs(gt_imap.importance[i] - max_importance) < 1e-5:
                    return cur_duration
    assert False, "Full summary must cover the best segment"
    return cur_duration


def metric_prf(gt_imap, pred_imap):
    """ Compute precision, recall and f-score.
    Precision is the number of detected segments matching at least 1 g.t. segment.
    Recall is the number of g.t. segments matching at least one detected.
    F-score is 2(p*r)/(p+r)
    """
    
    pred_matches = np.zeros(len(pred_imap), dtype=int)
    gt_matches = np.zeros(len(gt_imap), dtype=int)
    gt_periods = gt_imap.get_periods()
    pred_periods = pred_imap.get_periods()

    beta = 0.75  # another meaning than alpha
   
    for ip, ps in enumerate(pred_imap.segments):
        for ig, gs in enumerate(gt_imap.segments):
            assert(0 < gt_periods[ig] <= gs.length()), "Mistake in annotation"
            assert(0 < pred_periods[ip] <= ps.length()), "Mistake in annotation"
            
            inter = ps.intersection(gs)
            if inter.length() > 0:
                il = inter.length()
                pg = gt_periods[ig]
                pp = pred_periods[ip]
                if pp >= il and pg >= il:
                    ul = pp + pg - il
                elif pp >= il and pg <= il:
                    ul = pp
                elif pp <= il and pg >= il:
                    ul = pg
                else:
                    assert(pp < il and pg < il)
                    ul = il

                if il >  beta * ul:
                    pred_matches[ip] = 1
                    gt_matches[ig] = 1
    n_p = max(len(pred_imap), 1)
    n_g = max(len(gt_imap), 1)
    precision = sum(pred_matches) / float(n_p)
    recall = sum(gt_matches) / float(n_g)
    if (precision == 0 and recall == 0):
        f_score = 0.0
    else:
        f_score = 2 * precision * recall / (precision + recall)
    return precision, recall, f_score

    
