import os, sys
# Add the directory above to path
sys.path.insert(0, os.path.split(__file__)[0] + "/..")
from importance_map import Segment, EmptySegment, ImportanceMap, NoAnnotation
from testing import assert_eq_results as assert_eq
from testing import tic
import numpy as np

def test_Segment():
    S = Segment
    ES = EmptySegment

    x = S(-1, 1)
    y = S(0, 2)
    z = S(2, 4)

    # Length tests
    assert( x.length() == 2 )
    assert( y.length() == 2 )
    assert( z.length() == 2 )

    # Intersection tests
    assert( x.intersection(y) == S(0, 1) )
    assert( x.intersection(z) == ES() )
    assert( y.intersection(z) == S(2, 2) )

    for s in (x, y, z):
        assert( s.intersection(s) == s )
        
    # Union_len tests
    assert( x.union_length(y) == 3 )
    assert( x.union_length(z) == 4 )
    assert( y.union_length(z) == 4 )

    # Empty tests
    assert( not x.isempty() )
    assert( not y.isempty() )
    assert( not z.isempty() )
    assert( S(1,1).isempty() )
    assert( ES().isempty() )

    print "All 'Segment' tests passed"

def assert_imaps_equal(x, x_orig):
    for f in ('segments', 'importance', 'shot_tags', 'video_tags', 
        'vidname'):
        assert(getattr(x, f) == getattr(x_orig, f))

def test_ImportanceMap():
    IM = ImportanceMap
    S = Segment
    
    def fixture_x():
        x = IM("aaa", ends=[1, 10])
        x.video_tags = "prop:fps=1"
        x.importance = [2./3, 1./3]
        return x
    
    def fixture_y():
        y = IM("bbb", ends=[2, 4])  # maps to 1, 2 with fps 1
        y.video_tags = "prop:fps=2"
        y.importance = [1./3, 1./3]
        return y

    def test_append():
        x = fixture_x()
        x.append(fixture_x())
        assert(x.segments == [S(0, 1), S(1, 10), S(10, 11), S(11, 20)])
        assert(x.importance == [2./3, 1./3, 2./3, 1./3])
        
        # Different fps
        
        x = fixture_x()
        x.append(fixture_y())
        assert(x.segments == [S(0, 1), S(1, 10), S(10, 11), S(11, 12)])
        assert(x.importance == [2./3, 1./3, 1./3, 1./3])
        assert(x.get_fps() == 1)
        print "append OK"
    test_append()
    
    def fixture_z():
        z = IM("aaa", ends=[2, 3, 5])
        z.importance = [2./3, 1./3, 3./3]
        return z
    
    def test_merge_smaller():
        x = fixture_z()
        x.merge_smaller(2, method='toleft')
        assert(x.segments == [S(0, 3), S(3, 5)])
        assert(x.importance == [2./3, 3./3])
        
        x = fixture_z()
        x.merge_smaller(2, method='toright')
        assert(x.segments == [S(0, 2), S(2, 5)])
        assert(x.importance == [2./3, 3./3])
        
        x =  IM("a", ends=[3,4,5,6,10,20])
        x.importance = [3,1,2,3,2,1]
        x.merge_smaller(2, method='toleft')
        assert(x.segments == [S(0,4), S(4,6), S(6,10), S(10,20)])
        assert(x.importance == [3, 2, 2, 1])

        print "merge_smaller OK"
    test_merge_smaller()
    
    def fixture_t():
        t = IM("aaa", ends=[1,4,5,6])
        t.importance = [2./3, 1./3, 3./3, 3./3]
        return t

    def test_enlarge_smaller():
        z = fixture_z()
        z.enlarge_smaller(2)
        assert(z.segments == [S(0, 2), S(2, 4), S(3, 5)])
        assert(z.importance == fixture_z().importance)
        
        t = fixture_t()
        t.enlarge_smaller(3)
        # coinciding segments in the end are removed!
        assert(t.segments == [S(0, 3), S(1, 4), S(3, 6)])
        assert(t.importance == fixture_t().importance[:-1])
        print "enlarge_smaller OK"
    test_enlarge_smaller()
    
    def test_decrease_importance():
        z = fixture_z()
        z.decrease_importance(2)
        assert(z.segments == fixture_z().segments)
        assert(z.importance == [2./3, 1./6, 3./3])
        
        t = fixture_t()
        t.decrease_importance(3)
        assert(t.segments == fixture_t().segments)
        assert(t.importance == [2./9, 1./3, 3./9, 3./9])
        print "decrase_importance OK"
    test_decrease_importance()

    def test_top_limited_duration_inds():
        z = fixture_z()   
        inds = z.top_limited_duration_inds(3)
        assert(inds == [2, 0])
        assert_imaps_equal(z, fixture_z())
        
        t = fixture_t()
        inds = t.top_limited_duration_inds(3)
        assert(inds == [2, 3, 0])
        assert_imaps_equal(t, fixture_t())
        print "top_limited_duration_inds OK"
    test_top_limited_duration_inds()

    print "All 'ImportanceMap' tests passed"

def test_metrics():
    import metrics as ms
    mti = ms.metric_total_importance
    mir = ms.metric_importance_ratio
    prf = ms.metric_prf
    msd = ms.metric_msd
    
    IM = ImportanceMap
    S = Segment
    
    def X():
        x = IM("aaa", ends=[1, 3, 4])
        x.importance = [1./3, 2./3, 3./3]
        x.video_tags = "prop:fps=1"
        return x
    
    def Y():
        y = IM("bbb", ends=[2, 4])
        y.importance = [1./3, 2./3]
        y.video_tags = "prop:fps=1"
        return y

    def Z():
        z = X()
        z.shot_tags[1] = "prop:minlen=1s"
        z.video_tags = "prop:fps=1"
        return z
        
    def test_metric_importance_ratio():   
        x = X()
        y = Y()
        imp, dur = mti(x, y, max_duration=2, n_top=None)
        assert_imaps_equal(x, X())
        assert_imaps_equal(y, Y())
        assert(imp == 3./3)
        assert(dur == 2)
        
        imp, dur = mir(x, y, max_duration=2, n_top=None)
        assert_imaps_equal(x, X())
        assert_imaps_equal(y, Y())
        # ideal: 3./3 (len=0.8) + 2./3 (len=1.2 < 1.6 => does not cover g.t.)
        assert(imp == 1.0)
        assert(dur == 2)
        
        z = Z()
        imp, dur = mti(z, y, max_duration=2, n_top=None)
        assert_imaps_equal(z, Z())
        assert_imaps_equal(y, Y())
        assert(imp == 2./3 + 3./3)
        assert(dur == 2)
        
        imp, dur = mir(z, y, max_duration=2, n_top=None)
        assert_imaps_equal(z, Z())
        assert_imaps_equal(y, Y())
        # ideal: 3./3 (len=0.8) + 2./3 (len=0.8) +
        #   + 1./3 (len=0.4 => does not cover g.t.)
        assert(imp == 1.0)
        assert(dur == 2)
        print "metric_importance_ratio OK"
    test_metric_importance_ratio()

    def test_metric_msd():   
        # X-Y
        x = X()
        y = Y()
        dur = msd(x, y)
        assert_imaps_equal(x, X())
        assert_imaps_equal(y, Y())
        assert(dur == 2)

        y.video_tags = "prop:fps=2" # 2 fps => 2 frames = 1 sec 
        assert(msd(x, y) == 1)
        del y

        # Z-Y
        z = Z()
        y = Y()
        dur = msd(z, y)
        assert_imaps_equal(z, Z())
        assert_imaps_equal(y, Y())
        assert(dur == 2)
        
        # Y-X
        dur = msd(y, x)
        assert_imaps_equal(x, X())
        assert_imaps_equal(y, Y())
        assert(dur == 3)

        x.video_tags = "prop:fps=2"
        assert(msd(y, x) == 1.5)
        del x

        print "metric_msd OK"
    test_metric_msd()
        
    def test_metric_prf():
        x = X()
        y = Y()
        p, r, f = prf(x, y)
        assert(p == r == f == 0.0)
        assert_imaps_equal(x, X())
        assert_imaps_equal(y, Y())

        z = Z()
        p, r, f = prf(x, z)
        assert(p == r == f == 1.0)
        assert_imaps_equal(x, X())
        assert_imaps_equal(z, Z())

        p, r, f = prf(y, z)
        assert(p == r == f == 0.0)
        assert_imaps_equal(z, Z())
        assert_imaps_equal(y, Y())

        def T():
            t = IM("ttt", ends=[3, 4])
            return t
        t = T()
        p, r, f = prf(x, t)
        assert(np.allclose([p,r,f], [0.5, 1.0/3, 0.4]))
        assert_imaps_equal(x, X())
        assert_imaps_equal(t, T())

        def U1():
            u = IM("uuu1", ends=[4, 5])
            return u
        def U2():
            u = IM("uuu2", ends=[5])
            return u
        u1 = U1()
        u2 = U2()
        p, r, f = prf(u1, u2)
        assert(np.allclose([p,r,f], [1.0, 0.5, 2.0/3]))
        assert_imaps_equal(u1, U1())
        assert_imaps_equal(u2, U2())
        print "metric_prf OK"
    test_metric_prf()

    print "All metrics tests passed"


if __name__ == '__main__':

    all_tests = ['Segment', 'ImportanceMap', 'metrics']
    tests = []
    module = None
    if len(sys.argv) > 1:
        args = sys.argv[1:]
        while args:
            a = args.pop(0)
            if a in all_tests:
                tests.append(a)
            elif a == '-m':
                module = args.pop(0)
            else:
                raise KeyError("Unknown test: %s" % a)

    if tests == []:    
        tests = all_tests

    # Run all tests
    for t in tests:
        try:
            exec("test_%s()" % t)
        except:
            from traceback import print_exc
            print_exc(limit=5)
            print "ERROR in test %s" % t


    print "All tests finished."
